#include "pdcids.h"
#include "pdprocessor.h"

#include "public.sdk/source/vst/vstaudioprocessoralgo.h"
#include "public.sdk/source/vst/vsthelpers.h"
#include "base/source/fstreamer.h"
#include "pluginterfaces/vst/ivstparameterchanges.h"
#include <sys/stat.h>
#include <dlfcn.h>

#if SMTG_OS_MACOS || SMTG_OS_WINDOWS
extern char gPath[];
#endif

using namespace Steinberg;

namespace Steinberg {

    /* pdassoc - mechanism for associating pd instances with "processors" */
typedef struct _pdassoc
{
    PdProcessor *pda_proc;
    t_pdinstance *pda_pd;
} t_pdassoc;

static t_pdassoc *pda_vec = 0;
int pda_n = 0;

static void pda_add(PdProcessor *proc, t_pdinstance *pd)
{
    if (!pda_n)
        pda_vec = (t_pdassoc *)malloc(sizeof(*pda_vec));
    else pda_vec = (t_pdassoc *)realloc(pda_vec,
        (pda_n+1) * sizeof(*pda_vec));
    pda_vec[pda_n].pda_proc = proc;
    pda_vec[pda_n].pda_pd = pd;
    pda_n++;
    fprintf(stderr, "added instance %d\n", pda_n);
}

static void pda_rm(PdProcessor *proc)
{
    int i;
    for (i = 0; i < pda_n; i++)
        if (pda_vec[i].pda_proc == proc)
    {
        int nmove = pda_n - i - 1;
        if (nmove > 0)
            memmove(pda_vec+i, pda_vec+(i+1), nmove * sizeof(*pda_vec));
        if (pda_n <= 1)
        {
            free(pda_vec);
            pda_vec = 0;
        }
        else pda_vec = (t_pdassoc *)realloc(pda_vec,
             (pda_n-1) * sizeof(*pda_vec));
        pda_n--;
        fprintf(stderr, "remove instance, %d left\n", pda_n);
        return;
    }
    fprintf(stderr, "couldn't find instance to remove\n");
}

static PdProcessor *pda_getprocessor(t_pdinstance *pd)
{
    int i;
    if (pd)
        for (i = 0; i < pda_n; i++)
            if (pda_vec[i].pda_pd == pd)
                return (pda_vec[i].pda_proc);
    fprintf(stderr, "pd: can't find current VST plugin\n");
    return (0);
}

    /* callbacks from Pd */
void pdsetfile(t_pdinstance *pd, const char *name, const char *dir)
{    
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (proc)
        proc->setfilename(name, dir);
}

void pdsetparamvalue(t_pdinstance *pd, int paramindex, float f)
{
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (!proc)
        return;
    if (f < 0)
        f = 0;
    else if (f > 1)
        f = 1;
    /* fprintf(stderr, "param set %d %f\n", paramindex, f); */
    if (paramindex < 0 || paramindex >= PDNPARAMS+1)
        fprintf(stderr, "param number out of range\n");
    else if (f != proc->pd_paramvec[paramindex].pdp_f)
    {
        proc->pd_paramvec[paramindex].pdp_f = f;
        proc->pd_paramvec[paramindex].pdp_dirty = 1;
    }
}

void pdmidinote(t_pdinstance *pd, int pitch, int vel)
{
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    Vst::Event outevent;
    if (!proc || !proc->datap || !proc->datap->outputEvents)
    {
        fprintf(stderr, "midinote: bad instance ptrs\n");
        return;
    }
    outevent.type = Vst::Event::kNoteOnEvent;
    outevent.busIndex = 0;
    outevent.sampleOffset = 0;
    outevent.ppqPosition = 0;
    outevent.noteOn.channel = 0;
    outevent.noteOn.pitch = pitch;
    outevent.noteOn.tuning = 0;
    outevent.noteOn.velocity = (vel > 127 ? 1 : (vel < 0 ? 0 : vel/127.));
    outevent.noteOn.length = 0;
    outevent.noteOn.noteId = -1;
    proc->datap->outputEvents->addEvent(outevent);
}

void pdsetbuslist(t_pdinstance *pd, int inout, int argc, t_atom *argv)
{
    int count[PDMAXBUS], *nbusp, *busvec, destroy = 0, i;
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (!proc)
        return;
    fprintf(stderr, "set %d buses of type %d ...\n", argc, inout);
    for (i = 0; i < argc; i++)
    {
        count[i] = atom_getfloatarg(i, argc, argv);
        if (!count[i])
            count[i] = 2;
        count[i] += (count[i] & 1);
        if (count[i] < 2 || count[i] > 8)
            fprintf(stderr, "setting channel count %d to 8\n", count[i]),
                count[i] = 8;
    }
    if (inout)
        nbusp = &proc->pd_noutbus, busvec = proc->pd_outbusvec;
    else nbusp = &proc->pd_ninbus, busvec = proc->pd_inbusvec;
    if (argc < *nbusp)
        destroy = 1;
    for (i = 0; i < *nbusp; i++)
        if (busvec[i] != count[i])
            destroy = 1;
    if (destroy)
    {
        fprintf(stderr, "deleting existing buses\n");
        for (i = *nbusp; --i; )
        {
            Vst::AudioBus *ab =
                (inout? proc->getAudioOutput(i) : proc->getAudioInput(i));    
            if (ab)
                delete (ab);
            else fprintf(stderr, "couldn't delete bus %d\n", i);
        }
        *nbusp = 0;
    }
    for (i = *nbusp; i < argc; i++)
    {
         Vst::SpeakerArrangement speakers =
            (count[i] <= 2 ? Vst::SpeakerArr::kStereo :
                (count[i] <= 4 ? Vst::SpeakerArr::k40Music :
                (count[i] <= 6 ? Vst::SpeakerArr::k60Music :
                Vst::SpeakerArr::k80Music)));
         busvec[i] = count[i];
         if (inout)
         {
            fprintf(stderr, "new bus %d\n", count[i]);
            if (!proc->addAudioOutput(STR16("BUS"), speakers))
                fprintf(stderr, "... returned error\n");
         }
         else proc->addAudioInput(STR16("BUS"), speakers);
    }
    *nbusp = argc;
}

void pddefineparam(t_pdinstance *pd, int argc, t_atom *argv)
{
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (proc)
        proc->defineparam(argc, argv);
}


void pdquit(t_pdinstance *pd)
{
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (proc)
        proc->quit();
}

void pd_verifyvstquit( void)
{
    fprintf(stderr, "verifyquit");
    t_pdinstance *pd = libpd_this_instance();
    if (!pd)
    {
        fprintf(stderr, "no pd instance?\n");
        return;
    }
    PdProcessor *proc = static_cast<PdProcessor *>(pda_getprocessor(pd));
    if (proc)
        proc->quit();
    
}

PdProcessor::PdProcessor ()
{
    int i;
    setControllerClass (kPdControllerUID);
    processContextRequirements.needTempo ();

    libpd_init();
    pd1 = libpd_new_instance();
    pda_add(this, pd1);
    libpd_set_messagehook(pdmessage);
    libpd_set_instance(pd1);
    libpd_init_audio(PDDEFNCHAN, PDDEFNCHAN, 48000);
    libpd_start_message(1);
    libpd_add_float(1.0f);
    libpd_finish_message("pd", "dsp");
    pd_fromguihook = libpd_bind("pd");
    pd_frompdhook = libpd_bind("host");
    class_addmethod(glob_pdobject, (t_method)pd_verifyvstquit,
        gensym("verifyquit"), A_DEFFLOAT, 0);
    
    for (i = 0; i < PDNPARAMS+1; i++)
        pd_paramvec[i].pdp_f = 0, pd_paramvec[i].pdp_dirty = 0;
}

PdProcessor::~PdProcessor ()
{
    pda_rm(this);
    libpd_set_instance(pd1);
    if (pd_vis)
    {
        libpd_stop_gui();
        pd_vis = 0;
    }
    libpd_free_instance(pd1);
}

tresult PLUGIN_API PdProcessor::initialize (FUnknown* context)
{
    int i;
        /* initialize superclass */
    tresult result = AudioEffect::initialize (context);

    if (result != kResultOk)
    {
        return result;
    }
#ifdef __linux__
    if (!(pd_dlopened = dlopen(
        "/home/msp/.vst3/PureVST.vst3/Contents/x86_64-linux/PureVST.so",
            RTLD_GLOBAL | RTLD_NOW)))
    {
        fprintf(stderr, "couldn't reopen DLL\n");
    }
#else
    else fprintf(stderr, "linked OK\n");
    {
        char dllpath[MAXPDSTRING];
        strncpy(dllpath, gPath, MAXPDSTRING);
        strncat(dllpath, "/Contents/MacOS/PureVST",
            MAXPDSTRING-strlen(dllpath));
        dllpath[MAXPDSTRING-1] = 0;
        fprintf(stderr, "MACOS path: %s\n", dllpath);
        if (!(pd_dlopened = dlopen(dllpath, RTLD_GLOBAL | RTLD_NOW)))
            fprintf(stderr, "couldn't reopen DLL\n");
        else fprintf(stderr, "linked OK\n");
    }
#endif

        /* create Audio IO */
    addAudioInput(STR16 ("2ch 1 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 2 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 3 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 4 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 5 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 6 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 7 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 8 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 9 in"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioInput(STR16 ("2ch 10 in"),
        Steinberg::Vst::SpeakerArr::kStereo);

    addAudioOutput(STR16 ("Stereo Out 1"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 2"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 3"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 4"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 5"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 6"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 7"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 8"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 9"),
        Steinberg::Vst::SpeakerArr::kStereo);
    addAudioOutput(STR16 ("Stereo Out 10"),
        Steinberg::Vst::SpeakerArr::kStereo);
    pd_ninbus = 10;
    for (i = 0; i < pd_ninbus; i++)
        pd_inbusvec[i] = 2;
    pd_noutbus = 10;
    for (i = 0; i < pd_ninbus; i++)
        pd_outbusvec[i] = 2;

        /* create event bus (unused as yet) */
    addEventInput (STR16 ("Event In"), 1);
    addEventOutput (STR16 ("Event Out"), 1);
    fprintf(stderr, "init done\n");
    return kResultOk;
}

tresult PLUGIN_API PdProcessor::terminate ()
{
    return AudioEffect::terminate ();
}

tresult PLUGIN_API PdProcessor::setActive (TBool state)
{
        /* called when the Plug-in is enabled or disabled */
    return AudioEffect::setActive (state);
}

tresult PLUGIN_API PdProcessor::setBusArrangements
    (Vst::SpeakerArrangement* inputs, int32 numIns,
        Vst::SpeakerArrangement* outputs, int32 numOuts)
{
    auto *bus = FCast<Vst::AudioBus> (audioInputs.at (0));
    fprintf(stderr, "setBusArrangements xxx %d %d\n", numIns, numOuts);
    if (numOuts)
        fprintf(stderr, "out 0: %d\n",
            Vst::SpeakerArr::getChannelCount(outputs[0]));
        
    if (!bus)
        return (kResultFalse);
    getAudioInput (0)->setArrangement (Vst::SpeakerArr::kStereo);
    getAudioInput (0)->setName (STR16 ("Stereo In"));
    getAudioOutput (0)->setArrangement (Vst::SpeakerArr::kStereo);
    getAudioOutput (0)->setName (STR16 ("Stereo Out"));
    return (kResultTrue);
}

tresult PLUGIN_API PdProcessor::canProcessSampleSize (int32 symbolicSampleSize)
{
    if (symbolicSampleSize == Vst::kSample32
        /* || symbolicSampleSize == Vst::kSample64 */
            )
            return kResultTrue;
    else return kResultFalse;
}

static int find_pd_dir(char *buf)
{
    struct stat statbuf;
    const char *homedir = (getenv("HOME") ? getenv("HOME") : ".");
#if __linux__
    snprintf(buf, MAXPDSTRING-1, "%s/pd", homedir);
    buf[MAXPDSTRING-1] = 0;
    if (stat(buf, &statbuf) >= 0)
        return (1);
#else
    snprintf(buf, MAXPDSTRING-1,
        "%s/Applications/Pd.app/Contents/Resources", homedir);
    buf[MAXPDSTRING-1] = 0;
    if (stat(buf, &statbuf) >= 0)
        return (1);

    snprintf(buf, MAXPDSTRING-1,
        "/Applications/Pd.app/Contents/Resources");
    buf[MAXPDSTRING-1] = 0;
    if (stat(buf, &statbuf) >= 0)
        return (1);

#endif
    return (0);
}

#define PDBLOCK 64

tresult PLUGIN_API PdProcessor::process (Vst::ProcessData& data)
{
    //fprintf(stderr, "a"); fflush(stderr);
        /* get parameter changes */
    libpd_set_instance(pd1);
    datap = &data;
    if (data.inputEvents)
    {
        int32 n = data.inputEvents->getEventCount();
        Vst::Event event;
        for (int32 i = 0; i < n; i++)
            if (data.inputEvents->getEvent(i, event) == kResultOk)
        {
            switch (event.type)
            {
            case Vst::Event::kNoteOnEvent:
                libpd_start_message(2);
                libpd_add_float(event.noteOn.pitch);
                libpd_add_float(127 * event.noteOn.velocity);
                libpd_finish_list("midinote");
                break;
            case Vst::Event::kNoteOffEvent:
                libpd_start_message(2);
                libpd_add_float(event.noteOff.pitch);
                libpd_add_float(0);
                libpd_finish_list("midinote");
                break;
            }
        }
    }
    if (data.inputParameterChanges)
    {
        int32 numParamsChanged =
            data.inputParameterChanges->getParameterCount ();
        for (int32 index = 0; index < numParamsChanged; index++)
        {
            Vst::IParamValueQueue* paramQueue =
                data.inputParameterChanges->getParameterData (index);
            if (paramQueue)
            {
                Vst::ParamValue value;
                int32 sampleOffset;
                int32 numPoints = paramQueue->getPointCount();
                int32 paramid = paramQueue->getParameterId();
                
                if (paramid == PdParams::kParam1000Id)
                    paramid = PdParams::kParam1Id + PDNPARAMS;

                if (paramid >= PdParams::kParam1Id &&
                    paramid < PdParams::kParam1Id + (PDNPARAMS+1) &&
                        paramQueue->getPoint (numPoints - 1,
                            sampleOffset, value) == kResultTrue)
                {
                    /* fprintf(stderr, "param %d %f\n",
                        paramid - PdParams::kParam1Id, (float)value); */
                    int32 paramindex = paramid - PdParams::kParam1Id;
                    pd_paramvec[paramindex].pdp_dirty = 0;
                    pd_paramvec[paramindex].pdp_f = value;
                    libpd_set_instance(pd1);
                    libpd_start_message(2);
                    libpd_add_float(paramid - PdParams::kParam1Id);
                    libpd_add_float(value);
                    libpd_finish_list("param");
                }
            }
        }
    }

    pd_tickssincetempo++;
    if (data.processContext &&
        data.processContext->state & 
            (Vst::ProcessContext::kTempoValid |
                Vst::ProcessContext::kProjectTimeMusicValid))
    {
        double tempo =
            (data.processContext->state & Vst::ProcessContext::kTempoValid ?
                data.processContext->tempo : -1);
        Vst::TQuarterNotes bar =
            (data.processContext->state &
                Vst::ProcessContext::kBarPositionValid ?
                    data.processContext->projectTimeMusic : -1);
        if (tempo >= 0)
        {
            /* fprintf(stderr, "bar/tempo %f %f\n", bar, tempo); */
            libpd_start_message(2);
            libpd_add_float(bar);
            libpd_add_float(tempo);
            libpd_finish_list("transport");
            
            pd_tickssincetempo = 0;
        }
        pd_lasttempo = tempo;
    }
    
        /* audio */
    if (data.symbolicSampleSize == Vst::kSample32)
    {
        if (data.symbolicSampleSize == Vst::kSample32)
        {
            int nsamp;
            for (nsamp = 0; nsamp+PDBLOCK <= data.numSamples; nsamp += PDBLOCK)
            {
                int nbus, ch, cumch, i;
                float inbuf[PDDEFNCHAN*PDBLOCK], outbuf[PDDEFNCHAN*PDBLOCK];
                for (nbus = cumch = 0; nbus < data.numInputs; nbus++)
                    for (ch = 0; ch < data.inputs[nbus].numChannels &&
                        cumch < PDDEFNCHAN; ch++)
                {
                    if (data.symbolicSampleSize == Vst::kSample32)
                        for (i = 0; i < PDBLOCK; i++)
                    {
                        inbuf[cumch*PDBLOCK + i] = 
                            ((Vst::Sample32**)(getChannelBuffersPointer
                                (processSetup,
                                    data.inputs[nbus])))[ch][nsamp+i];
                    }
                    else for (i = 0; i < PDBLOCK; i++)
                    {
                        inbuf[cumch*PDBLOCK + i] = 
                            ((Vst::Sample64**)(getChannelBuffersPointer
                                (processSetup,
                                    data.inputs[nbus])))[ch][nsamp+i];
                    }
                    cumch++;
                }
                while (cumch < PDDEFNCHAN)
                {
                    for (i = 0; i < PDBLOCK; i++)
                        inbuf[ch*PDBLOCK + i] = 0;
                    cumch++;
                }
                libpd_set_instance(pd1);
                libpd_process_raw(inbuf, outbuf);
                for (nbus = cumch = 0; nbus < data.numOutputs; nbus++)
                    for (ch = 0; ch < data.outputs[nbus].numChannels &&
                        cumch < PDDEFNCHAN; ch++)
                {
                    if (data.symbolicSampleSize == Vst::kSample32)
                        for (i = 0; i < PDBLOCK; i++)
                    {
                        ((Vst::Sample32**)(getChannelBuffersPointer
                            (processSetup, data.outputs[nbus])))[ch][nsamp+i] =
                                outbuf[cumch*PDBLOCK + i];
                    }
                    else for (i = 0; i < PDBLOCK; i++)
                    {
                        ((Vst::Sample64**)(getChannelBuffersPointer
                            (processSetup, data.outputs[nbus])))[ch][nsamp+i] =
                                outbuf[cumch*PDBLOCK + i];
                    }
                    cumch++;
                }
                while (nbus < data.numOutputs)
                {
                    while (ch < data.outputs[nbus].numChannels)
                    {
                        if (data.symbolicSampleSize == Vst::kSample32)
                            for (i = 0; i < PDBLOCK; i++)
                        {
                            ((Vst::Sample32**)(getChannelBuffersPointer
                                (processSetup, data.outputs[nbus])))
                                    [ch][nsamp+i] = 0;
                        }
                        else for (i = 0; i < PDBLOCK; i++)
                        {
                            ((Vst::Sample32**)(getChannelBuffersPointer
                                (processSetup, data.outputs[nbus])))
                                    [ch][nsamp+i] = 0;
                        }
                        ch++;
                     }
                     ch = 0;
                     nbus++;
                }
            }
        }
        else fprintf(stderr, "unknown sample size");
    }

    //fprintf(stderr, "b"); fflush(stderr);
    if (data.outputParameterChanges)
    {
        int i;
        for (i = 0; i < PDNPARAMS+1; i++)
            if (pd_paramvec[i].pdp_dirty)
        {
            int shit1 = 0, shit2 = 0;
            Vst::IParamValueQueue *paramQueue =
                data.outputParameterChanges->addParameterData (
                    (i < PDNPARAMS ? kParam1Id+i : kParam1000Id), shit1);
            if (paramQueue)
                 paramQueue->addPoint (0, pd_paramvec[i].pdp_f, shit2);
            pd_paramvec[i].pdp_dirty = 0;
        }
    }
    if (pd_vis && pd_paramvec[PDNPARAMS].pdp_f < 0.5)
    {
        libpd_stop_gui();
        pd_vis = 0;
    }
    else if (!pd_vis && pd_paramvec[PDNPARAMS].pdp_f >= 0.5)
    {
        char pddir[MAXPDSTRING];
        if (find_pd_dir(pddir))
        {
            fprintf(stderr, "start gui: %s\n", pddir);
            libpd_start_gui(pddir);
        }
        else fprintf(stderr, "couldn't find Pd gui:\n");
        pd_vis = 1;
    }
    //fprintf(stderr, "c"); fflush(stderr);
    return kResultOk;
}

tresult PLUGIN_API PdProcessor::setupProcessing (Vst::ProcessSetup& newSetup)
{
        /* called before any processing */
    return AudioEffect::setupProcessing (newSetup);
}

tresult PLUGIN_API PdProcessor::getState (IBStream* state)  /* save state */
{
    IBStreamer streamer (state, kLittleEndian);
    streamer.writeStr8 (pd_filename);
    streamer.writeStr8 (pd_filedir);

    return kResultOk;
}

tresult PLUGIN_API PdProcessor::setState (IBStream* state) /* restore state */
{
    const char8 *filename, *filedir;

    if (!state)
        return kResultFalse;

    IBStreamer streamer(state, kLittleEndian);

    filename = streamer.readStr8();
    filedir = streamer.readStr8();

    if (!pd_open)
    {
        libpd_openfile(filename, filedir);
        fprintf(stderr, "opened patch: newest=%p\n", pd_newest);
        pd_open = 1;
    }

    fprintf(stderr, "setstate %s %s\n", filename, filedir);
    strncpy(pd_filename, filename, MAXPDSTRING-1);
    pd_filename[MAXPDSTRING-1] = 0;
    strncpy(pd_filedir, filedir, MAXPDSTRING-1);
    pd_filename[MAXPDSTRING-1] = 0;

    fprintf(stderr, "filename = %s, dir = %s\n", pd_filename, pd_filedir);
    /* delete filename;
    delete filedir; */ /* memory leak... */
    fprintf(stderr, "setstate done\n");
    return kResultOk;
}

void PdProcessor::setfilename(const char *name, const char *dir)
{
    strncpy(pd_filename, name, MAXPDSTRING - 1);
    pd_filename[MAXPDSTRING - 1] = 0;
    strncpy(pd_filedir, dir, MAXPDSTRING - 1);
    pd_filedir[MAXPDSTRING - 1] = 0;
}

void PdProcessor::defineparam(int argc, t_atom *argv)
{
    char str[80];
    snprintf(str, 79, "%d %f %f %s",
        (int)atom_getfloatarg(0, argc, argv),
        atom_getfloatarg(1, argc, argv),
        atom_getfloatarg(2, argc, argv),
        atom_getsymbolarg(3, argc, argv)->s_name);
    str[79] = 0;
    /* fprintf(stderr, "msg %s\n", str); */
    Vst::IMessage *msg = allocateMessage();
    msg->setMessageID("param");
    msg->getAttributes()->setBinary("Text", str, 80);
    sendMessage(msg);
    msg->release();
}

void PdProcessor::quit( void)
{
    pd_paramvec[PDNPARAMS].pdp_f = 0;
    pd_paramvec[PDNPARAMS].pdp_dirty = 1;
}

tresult PLUGIN_API PdProcessor::notify(Vst::IMessage *msg)
{
    if (!strcmp(msg->getMessageID(), "open"))
    {
        fprintf(stderr, "got open msg\n");
        if (!pd_vis)
        {
            char pddir[MAXPDSTRING];
            libpd_set_instance(pd1);
            if (find_pd_dir(pddir))
            {
                fprintf(stderr, "start gui: %s\n", pddir);
                libpd_start_gui(pddir);
            }
            else fprintf(stderr, "couldn't find Pd gui:\n");
            pd_paramvec[PDNPARAMS].pdp_f = 1;
            pd_paramvec[PDNPARAMS].pdp_dirty = 1;
            pd_vis = 1;
        }
        fprintf(stderr, "done open message\n");
    }
    else
    {
        fprintf(stderr, "unknown notifier message\n");
        return (kResultFalse);
    }
    return (kResultTrue);
}

} /* namespace Steinberg */


extern "C" {
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static void pdmessage(const char *recv, const char *msg, int argc, t_atom *argv)
{
    int i;
#if 0
    fprintf(stderr, "msg: %s %s", recv, msg);
    for (i = 0; i < argc; i++)
    {
        t_atom *a = &argv[i];
        if (libpd_is_float(a))
            fprintf(stderr, " %f", libpd_get_float(a));
        else if (libpd_is_symbol(a))
             fprintf(stderr, " %s", libpd_get_symbol(a));
    }
    fprintf(stderr, "\n");
#endif
    if (!strcmp(recv, "pd"))
    {
        if (!strcmp(msg, "open") && argc >= 2 &&
            libpd_is_symbol(argv) && libpd_is_symbol(argv+1))
                pdsetfile(libpd_this_instance(),
                    libpd_get_symbol(argv), libpd_get_symbol(argv+1));  

    }
    else if (!strcmp(recv, "host"))
    {
        if (!strcmp(msg, "param") && argc >= 2 &&
            libpd_is_float(argv) && libpd_is_float(argv+1))
                pdsetparamvalue(libpd_this_instance(),
                    (int)(libpd_get_float(argv)), libpd_get_float(argv+1));
        else if (!strcmp(msg, "midinote") && argc >= 2 &&
            libpd_is_float(argv) && libpd_is_float(argv+1))
                pdmidinote(libpd_this_instance(),
                    (int)(libpd_get_float(argv)), (int)libpd_get_float(argv+1));
        else if (!strcmp(msg, "in-bus-list"))
            pdsetbuslist(libpd_this_instance(), 0, argc, argv);
        else if (!strcmp(msg, "out-bus-list"))
            pdsetbuslist(libpd_this_instance(), 1, argc, argv);
        else if (!strcmp(msg, "define-param"))
            pddefineparam(libpd_this_instance(), argc, argv);
        else if (!strcmp(msg, "verifyquit"))
            pdquit(libpd_this_instance());
       else fprintf(stderr, "unknown or short message 'host %s'\n", msg);
    }
    else fprintf(stderr, "unknown msg %s %s (%d args)\n",
        recv, msg, argc);
}


t_symbol s_pointer, s_float, s_symbol, s_bang, s_list, s_anything,
  s_signal, s__N, s__X, s_x, s_y, s_;

void vst_cleanser(t_symbol **s)
{
    if (*s >= &s_pointer && *s <= &s_)
    {
        fprintf(stderr, "replace symbol '%s'\n", (*s)->s_name);
        *s = gensym((*s)->s_name);
    }
}


} /* extern "C" */

